import regex

name_regex = regex.compile(r'^[a-zA-Z0-9-]+$')

questions = [
    {
        "type": "input",
        "name": "name",
        "validate": lambda x: bool(name_regex.match(x)),
        "message": "Function name"
    },
    {
        "type": "input",
        "name": "description",
        "message": "Description of the function"
    },
    {
        "type": "list",
        "name": "runtime",
        "choices": [
            "nodejs12.x", "nodejs10.x",
            "python3.8", "python3.7", "python3.6", "python2.7",
            "ruby2.5",
            "java11", "java8",
            "dotnetcore2.1",
            "go1.x"
        ],
        "message": "Function runtime"
    },
    {
        "type": "input",
        "name": "timeout",
        "validate": lambda x: x.isnumeric(),
        "filter": lambda x: int(x),
        "message": "Function timeout (sec)"
    },
    {
        "type": "input",
        "name": "handler",
        "validate": lambda x: '.' in x,
        "message": "Function handler"
    }
]



questions = [
    {
        "type": "input",
        "name": "access_id",
        "message": "AWS access id"
    },
    {
        "type": "input",
        "name": "secret_key",
        "message": "AWS secret key"
    }
]



import boto3

from .function import init_lambda
from .iam import init_iam
from .s3 import init_s3

inits = [
    init_lambda,
    init_iam,
    init_s3,
]


class AWSProvider():
    creds = dict()
    lambda_role = None

    from .shared import _create_creds, _init_creds

    def __init__(self, global_config, options):
        self.gb = global_config
        self.go = options
        self._init_creds()
        for init in inits:
            init(self)

    from .function import create_function, delete_function, update_function
    from .layer import delete_layer, publish_layer
    from .s3 import _bucket_exist, _send_to_s3
    from .iam import _create_lambda_role



import boto3

def init_iam(self):
    self.iam = boto3.client(
        'iam',
        region_name=self.gb['region'],
        aws_access_key_id=self.creds['access_id'],
        aws_secret_access_key=self.creds['secret_key'],
    )

def _create_lambda_role(self):
    # TODO:
    return "arn:aws:iam::394115634019:role/dev-GraphQLRole-1EHEIW52JVMUS"



import regex

name_regex = regex.compile(r'^[a-zA-Z0-9-]+$')

questions = [
    {
        "type": "input",
        "name": "name",
        "validate": lambda x: bool(name_regex.match(x)),
        "message": "Project name"
    },
    {
        "type": "list",
        "name": "region",
        "choices": [
            "us-east-2",
            "us-east-1",
            "us-west-1",
            "us-west-2",
            "eu-central-1",
            "eu-west-1",
            "eu-west-2",
            "eu-west-3",
            "eu-north-1",
            "ap-east-1",
            "ap-south-1",
            "ap-northeast-3",
            "ap-northeast-2",
            "ap-southeast-1",
            "ap-southeast-2",
            "ap-northeast-1",
            "ca-central-1",
            "cn-north-1",
            "cn-northwest-1",
            "me-south-1",
            "sa-east-1",
            "us-gov-east-1",
            "us-gov-west-1"
        ],
        "message": "Project region"
    },
    {
        "type": "list",
        "name": "cloud",
        "choices": ["aws"],
        "message": "Cloud provider"
    }
]


import os
import boto3

def init_lambda(self):
    self.lb_client = boto3.client(
        'lambda',
        region_name=self.gb['region'],
        aws_access_key_id=self.creds['access_id'],
        aws_secret_access_key=self.creds['secret_key'],
    )


def delete_function(self, config):
    return self.lb_client.delete_function(
        FunctionName=config['name'],
    )


def update_function(self, config, stage):
    path = config.get('source', config['name'])
    # TODO: stage
    if not config.get('role'):
        config['role'] = self.lambda_role or self._create_lambda_role()
    if not os.path.exists(path):
        raise Exception(f"Error in {config['name']}, source do not exist!")
    else:
        code = self._send_to_s3(path)

    others_configs = {
        k: v for k, v in {
            'Description': config.get('description'),
            'Timeout': config.get('timeout'),
            'MemorySize': config.get('memory'),
            'VpcConfig': config.get('vpc'),
            'Layers': config.get('layers')
        }.items() if v
    }
    if config.get('environ'):
        others_configs['Environment'] = {
            'Variables': config['environ']
        }

    res = self.lb_client.update_function_code(
        FunctionName=config['name'],
        **code,
        Publish=(stage != self.gb['stages'][0]),
    )
    self.lb_client.update_function_configuration(
        FunctionName=config['name'],
        Runtime=config.get('runtime', 'provided'),
        Role=config['role'],
        Handler=config['handler'],
        **others_configs
    )
    if stage == self.gb['stages'][0]:
        return
    self.lb_client.update_alias(
        Name=stage,
        FunctionName=config['name'],
        FunctionVersion=res['Version']
    )


def create_function(self, config):
    path = config.get('source', config['name'])

    if not config.get('role'):
        config['role'] = self.lambda_role or self._create_lambda_role()
    if not os.path.exists(path):
        raise Exception(f"Error in {config['name']}, source do not exist!")
    else:
        code = self._send_to_s3(path)

    others_configs = {
        k: v for k, v in {
            'Description': config.get('description'),
            'Timeout': config.get('timeout'),
            'MemorySize': config.get('memory'),
            'VpcConfig': config.get('vpc'),
            'Layers': config.get('layers')
        }.items() if v
    }
    if config.get('environ'):
        others_configs['Environment'] = {
            'Variables': config['environ']
        }

    arn = self.lb_client.create_function(
        FunctionName=config['name'],
        Runtime=config.get('runtime', 'provided'),
        Role=config['role'],
        Handler=config['handler'],
        Code=code,
        Publish=config.get('publish', False),
        **others_configs
    )['FunctionArn']

    for stage in self.gb['stages']:
        self.lb_client.create_alias(
            Name=stage,
            FunctionName=arn,
            FunctionVersion='$LATEST',
            Description=f'{stage} grade function'
        )


from dotenv import load_dotenv
import argparse

from .handlers import (
    init_app,
    push_app,
    create_app,
    delete_app,
)

parser = argparse.ArgumentParser('pylone')
parser.add_argument('--creds-path', '-c', metavar='PATH', type=str, help="Credential path", default=".creds")

# SUBPARSER CONFIG
subparser = parser.add_subparsers(
    dest='action', title='action', description='Pylone actions', required=True)

# INIT
init = subparser.add_parser('init', help='initialize a new project')
init.set_defaults(handler=init_app)

# HOST
host = subparser.add_parser('host', help='host project in the cloud')
host.set_defaults(handler=create_app)

# DELETE
delete = subparser.add_parser('delete', help='delete project from the cloud')
delete.set_defaults(handler=delete_app)

# PUSH
push = subparser.add_parser('push', help='push modifications to the cloud')
push.add_argument('--force-update', '-f', action='store_true', help='force project update', default=False)
push.add_argument('--stage', '-s', type=str, help='project stage', default='dev')
push.add_argument('projects', metavar='NAME', type=str, nargs='*', help='projects to push (all by default)')
push.set_defaults(handler=push_app)


def main():
    load_dotenv('.env')
    options = parser.parse_args()

    if options.handler:
        options.handler(options)



from .config import save_config
from .providers import providers
from .utils.dirs import makedirs
from .utils.scripts import run
from .questions import qload

from .functions import PyloneFct
from .layers import PyloneLayer
from .apis import PyloneApi


class PyloneProject():
    functions = list()
    layers = list()
    apis = list()

    def __init__(self, options, config):
        self.config = config
        self.options = options
        self.provider = providers[config['cloud']](config, options)
        self._init_classes()

    def _init_classes(self):
        gcf = {
            'provider': self.provider,
        }
        for cfg in self.config.get('functions', {}).values():
            self.functions.append(PyloneFct(cfg, gcf))
        for cfg in self.config.get('layers', {}).values():
            self.layers.append(PyloneLayer(cfg, gcf))
        for cfg in self.config.get('apis', {}).values():
            self.apis.append(PyloneApi(cfg, gcf))

    def create_archi(self):
        for elem in [*self.layers, *self.functions, *self.apis]:
            if elem.cf.get('before-script'):
                run(elem.cf['before-script'])
            elem.create()
            if elem.cf.get('after-script'):
                run(elem.cf['after-script'])
        print(f'🚀 {self.config["name"]} hosted successfully')

    def delete_archi(self):
        for elem in [*self.apis, *self.functions, *self.layers]:
            elem.remove()
        print(f'🤯 {self.config["name"]} removed successfully')

    def update(self, stage):
        for elem in [*self.layers, *self.functions, *self.apis]:
            if self.options.force_update or elem.check_for_update(stage):
                if elem.cf.get('before-script'):
                    run(elem.cf['before-script'])
                elem.update(stage)
                if elem.cf.get('after-script'):
                    run(elem.cf['after-script'])
        print(f'🦄 {self.config["name"]} updated successfully')



import os
import json
from PyInquirer import prompt
from . import (
    create_function,
    global_config,
    create_layer,
    credentials,
)

QUESTIONS = {
    'create_function': create_function.questions,
    'create_layer': create_layer.questions,
    'credentials': credentials.questions,
    'global_config': global_config.questions,
}


def ask(message, default=True):
    question = [
        {
            'type': 'confirm',
            'name': 'out',
            'message': message,
            'default': default
        }
    ]
    return prompt(question).get('out', default)


def qload(fct_name):
    '''
    Load and ask a question file.
    '''
    return prompt(QUESTIONS[fct_name])



class PyloneApi():
    def __init__(self, config, global_config):
        self.cfg = config
        self.gcf = global_config

    def check_for_update(self, stage):
        pass

    def update(self, stage):
        pass

    def create(self):
        pass

    def remove(self):
        pass


import yaml


def fct_template(config, dest):
	"""
	This function build the function config file.
	"""
	with open(dest, 'w+') as fp:
		yaml.dump({
			'name': config['name'],
			'timeout': config['timeout'],
			'handler': config['handler'],
			'runtime': config['runtime'],
			'role': config['role']
		}, fp)



from botocore.exceptions import ClientError


class PyloneFct():
    def __init__(self, config, global_config):
        self.cf = config
        self.provider = global_config['provider']

    def check_for_update(self, stage):
        # TODO:
        return True

    def update(self, stage):
        self.provider.update_function(self.cf, stage)

    def create(self):
        self.provider.create_function(self.cf)

    def remove(self):
        try:
            self.provider.delete_function(self.cf)
        except ClientError as e:
            if e.response['Error']['Code'] == 'ResourceNotFoundException':
                print('Ressource do not exist')
            else:
                raise



import os
import json
import boto3

from ...questions import qload
from ...utils.gitignore import add_to_gitignore


def _init_creds(self):
    if os.path.exists(self.go.creds_path):
        with open(self.go.creds_path) as fp:
            j = json.load(fp)

        if 'aws' in j:
            self.creds = j['aws']
        else:
            self._create_creds()
    else:
        self._create_creds()


def _create_creds(self):
    self.creds = qload('credentials')
    with open(self.go.creds_path, 'w+') as fp:
        json.dump({'aws': self.creds}, fp)
    add_to_gitignore(self.go.creds_path.split('/')[-1])



from os import path


def add_to_gitignore(filenames):
    if isinstance(filenames, str):
        filenames = [filenames]
    if not path.isfile('.gitignore'):
        return
    with open('.gitignore') as fp:
        file = fp.read().strip() + '\n\n# Pylone\n'
    for name in filenames:
        if name in file: continue
        file = file + f'{name}\n'
    with open('.gitignore', 'w') as fp:
        fp.write(file.strip() + '\n')



import os
import boto3
import shutil

from ...utils.checksum import md5_file
from botocore.client import ClientError


def init_s3(self):
    self.s3 = boto3.resource(
        's3',
        region_name=self.gb['region'],
        aws_access_key_id=self.creds['access_id'],
        aws_secret_access_key=self.creds['secret_key'],
    )


def _bucket_exist(self, bucket):
    try:
        self.s3.meta.client.head_bucket(Bucket=bucket)
    except ClientError:
        return False
    return True


def _send_to_s3(self, path):
    if '.zip' in path:
        shutil.move(path, '/tmp/update.zip')
    else:
        shutil.make_archive('/tmp/update', 'zip', path)
    key = md5_file('/tmp/update.zip')
    if not self._bucket_exist('pylone-bucket'):
        self.s3.create_bucket(
            Bucket='pylone-bucket',
            CreateBucketConfiguration={
                'LocationConstraint': self.gb['region']
            }
        )
    self.s3.meta.client.upload_file(
        '/tmp/update.zip',
        'pylone-bucket',
        key,
    )
    return {
        'S3Bucket': 'pylone-bucket',
        'S3Key': key
    }



import os

def publish_layer(self, config):
    path = config.get('source', config['name'])

    if not os.path.exists(path):
        raise Exception(f"Error in {config['name']}, source do not exist!")
    else:
        code = self._send_to_s3(path)

    others_configs = {
        k: v for k, v in {
            'Description': config.get('description'),
            'LicenseInfo': config.get('licenseInfo'),
            'MemorySize': config.get('memory'),
            'VpcConfig': config.get('vpc'),
            'Environment': config.get('environ'),
        }.items() if v
    }

    return self.lb_client.publish_layer_version(
        LayerName=config['name'],
        Content=code,
        CompatibleRuntimes=config['runtimes'],
        **others_configs
    )['LayerVersionArn']


def delete_layer(self, config):
    nx = 0
    layers = list()

    while nx or nx == 0:
        marker = {'Marker': nx} if nx else dict()
        res = self.lb_client.list_layer_versions(
            LayerName=config['name'],
            MaxItems=50,
            **marker
        )
        nx = res.get('NextMarker')
        layers += res.get('LayerVersions', [])

    for layer in layers:
        try:
            self.lb_client.delete_layer_version(
                LayerName=config['name'],
                VersionNumber=layer['Version']
            )
        except Exception as e:
            print('Error when deleting a version layer')
            print(e)



import os

from .config import get_global_config, create_global_config
from .utils.changes import detect_changes
from .pylone import PyloneProject
from .questions import ask


def init_app(options):
    config = get_global_config()
    if not config or not ask('A config file was detected would you use it'):
        config = create_global_config()

    project = PyloneProject(options, config)


def create_app(options):
    config = get_global_config()
    if not config:
        exit('No config file found!')
    project = PyloneProject(options, config)
    project.create_archi()

def delete_app(options):
    config = get_global_config()
    if not config:
        exit('No config file found!')
    project = PyloneProject(options, config)
    project.delete_archi()


def push_app(options):
    config = get_global_config()
    if not config:
        exit('No config file found!')
    project = PyloneProject(options, config)
    project.update(options.stage)
    # if not options.functions_only:
    #     project.push_layers()
    # if not options.layers_only:
    #     project.push_functions()



from setuptools import setup, find_namespace_packages

setup(
    name='pylone',
    version='0.1.7',
    description='A Python Serverless framework',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/mathix420/pylone',
    author='Arnaud Gissinger',
    author_email='agissing@student.42.fr',
    license='MIT',
    python_requires='>=3.6',
    classifiers=[
                'Intended Audience :: Developers',
                'Intended Audience :: System Administrators',

                'Topic :: Software Development :: Build Tools',

                'License :: OSI Approved :: MIT License',

                'Programming Language :: Python :: 3 :: Only',
                'Programming Language :: Python :: 3.6',
                'Programming Language :: Python :: 3.7',
                'Programming Language :: Python :: 3.8',
                'Programming Language :: Python :: 3.9',
    ],
    install_requires=[
        'python-dotenv==0.10.5',
        'PyInquirer==1.0.3',
        'PyYAML>=3.13',
        'boto3>=1.11.6',
    ],
    packages=find_namespace_packages(include=["src", "src.*"]),
    entry_points={'console_scripts': ['pylone=src.__main__:main']},
)



import os

def makedirs(directories):
	for directory in directories:
		try:
			os.mkdir(directory)
		except FileExistsError:
			print(directory, 'already exists')
		except Exception as e:
			print('Error when creating directories')
			print(e)


import yaml
import os
import re

from .questions import qload
from .utils.constructors import get_env

variables = re.compile(r'\s(\$[A-Z_]+)\s')
yaml.add_constructor('!env', get_env, yaml.SafeLoader)


def get_global_config():
    if not os.path.exists('pylone.yaml'):
        return None
    with open('pylone.yaml') as fp:
        config = yaml.load(fp, yaml.SafeLoader)
    return config


def create_global_config():
    config = qload('global_config')
    save_config(config)
    return config


def save_config(config):
    with open('./pylone.yaml', 'w+') as fp:
        yaml.dump(config, fp, default_flow_style=False, indent=2)



import subprocess


def detect_changes(path=None, ref_commit=None, cwd='.'):
    """
    This function return True if changes are detected in the `path` location
    """
    command = ['git', 'diff']
    if ref_commit:
        command.append(ref_commit)
    if path:
        command.append(path)
    process = subprocess.run(command, capture_output=True, cwd=cwd)
    return bool(process.stdout)


def get_last_commit(branch='HEAD', cwd='.'):
    """
    This function return the last commit hash
    """
    process = subprocess.run(
        ['git', 'rev-parse', branch],
        capture_output=True,
        cwd=cwd
    )
    # Removing the '\n'
    return process.stdout.decode('utf-8')[:-1]